import pickle
import torch.utils.data
from collections import defaultdict
import math
import hpc_hd_data.data
from transformers import BatchEncoding
import abc
import functools
from io import BytesIO
from typing import Any, Callable, Generic, Iterable, Literal, Sequence, TypeVar, TypedDict, cast
from typing_extensions import TypeVarTuple, Unpack
from hscitorchutil.fsspec import get_s3fs_credentials, cache_locally_if_remote
from importlib import resources
from hereutil import here
import pandas as pd
import torch
from transformers import AutoTokenizer, AutoModel
import torch.utils.data.dataloader
from PIL import Image
from more_itertools import chunked
import numpy as np
from torch.utils.data import Dataset
from stream_unzip import stream_unzip
from hscitorchutil.dataset import ABaseDataModule, identity_transformation, EntryTransformingMapDataset
from hscitorchutil.sqlite import SQLiteDataset
from hscitorchutil.fsspec import fetch_async, AFetcher, get_async_filesystem, get_s3fs_credentials, prefetch_if_remote, PLocalAFetcher

import logging

T_co = TypeVar('T_co', covariant=True)
T2_co = TypeVar('T2_co', covariant=True)
Ts = TypeVarTuple("Ts")


class ECCOIdBasedSubdataset(Dataset[tuple[Unpack[Ts]]], Generic[Unpack[Ts]]):
    def __init__(self,
                 ecco_ids: Sequence[str],
                 dataset: SQLiteDataset[tuple[Unpack[Ts]]],
                 ecco_id_column: str = 'ecco_id'
                 ):
        self.dataset = dataset
        self.expanded_ids = []
        for ecco_ids_c in chunked(ecco_ids, 1000):
            ecco_ids_s = '", "'.join(ecco_ids_c)
            self.expanded_ids.extend(map(lambda t: t[0], self.dataset.sqlite.execute(
                f'SELECT {self.dataset.index_column} FROM {self.dataset.table_name} WHERE {ecco_id_column} IN ("{ecco_ids_s}")').fetchall()))

    def __len__(self) -> int:
        return len(self.expanded_ids)

    def __getitem__(self, idx: int) -> tuple[Unpack[Ts]]:
        return self.dataset[self.expanded_ids[idx]]

    def __getitems__(self, idxs: Sequence[int]) -> Sequence[tuple[Unpack[Ts]]]:
        return self.dataset.__getitems__([self.expanded_ids[idx] for idx in idxs])


def _filter_out_nones(tuples: Iterable[tuple[Unpack[Ts]]]) -> Sequence[tuple[Unpack[Ts]]]:
    return list(filter(lambda x: not any(map(lambda x: x is None, x)), tuples))


def _uncompress(compressed_bytes: bytes) -> BytesIO:
    _, _, biter = next(stream_unzip([compressed_bytes]))
    bio = BytesIO()
    for b in biter:
        bio.write(b)
    return bio


def _load_and_transform_image(image_bytes: BytesIO, transform: Callable[[Image.Image], T_co]) -> T_co | None:
    try:
        im = Image.open(image_bytes)
        return transform(im)
    except:
        logging.exception("Failed to load/transform image, returning None")
        return None


def _load_from_zip_uncompress_and_transform(offsets_and_sizes: Sequence[tuple[int, int]], afetcher: AFetcher, transform: Callable[[BytesIO], T_co]) -> list[T_co]:
    try:
        return [transform(_uncompress(compressed_bytes)) for compressed_bytes in fetch_async(afetcher, [(offset, offset+size) for offset, size in offsets_and_sizes])]
    except Exception as e:
        logging.exception(
            "Failed to load data from zip, returning empty batch")
        return []


def _ecco_images_transform(offsets_sizes_and_ids: Sequence[tuple[str, int, int, Unpack[Ts]]], afetcher: AFetcher, image_transform: Callable[[Image.Image], T_co]) -> Sequence[tuple[str, T_co]]:
    return _filter_out_nones(
        zip(
            [id for id, *_ in offsets_sizes_and_ids],
            _load_from_zip_uncompress_and_transform(
                [(offset, size)
                 for _, offset, size, *_ in offsets_sizes_and_ids],
                afetcher=afetcher,
                transform=functools.partial(
                    _load_and_transform_image, transform=image_transform)  # type: ignore
            ),
        )
    )


def _ecco_images_ds_transform(ds: Dataset[tuple[str, int, int, Unpack[Ts]]], _: pd.DataFrame, afetcher: AFetcher, image_transform: Callable[[Image.Image], T_co]) -> Dataset[tuple[str, T_co]]:
    return EntryTransformingMapDataset(dataset=ds, transform=functools.partial(_ecco_images_transform, afetcher=afetcher, image_transform=image_transform))


def transform_binary_PIL_image_to_bool_tensor(img: Image.Image) -> torch.Tensor:
    """Convert a PIL Image to a torch bool tensor"""
    return (
        torch.from_numpy(np.array(img, dtype=bool))
        .view(img.size[1], img.size[0], 1)
        .permute((2, 0, 1))
        .contiguous()  # HWC to CHW
    )


def _load_multi_target_split() -> pd.DataFrame:
    return pd.read_csv(here("data/work/ecco-multi-target-split.tsv.gz").open('rb'),
                       compression='gzip',
                       sep="\t",
                       low_memory=False,
                       index_col='ecco_id',
                       dtype=dict(
                           ecco_id='string',
                           ecco_part='string',
                           ecco_module='string',
                           publication_year='Int16',
                           publication_year_definite='boolean',
                           earliest_publication_year='Int16',
                           earliest_publication_year_definite='boolean',
                           primary_language='string',
                           total_price='float',
                           publication_place='string',
                           publication_country='string',
                           publication_place_deduced='boolean',
                           pages='Int16',
                           estc_id='string',
                           set='string',
                           major_primary_language='string',
                           major_publication_place='string',
                           major_publication_country='string')
                       )


def _load_multi_target_actor_split() -> pd.DataFrame:
    return pd.read_csv(here("data/work/ecco-multi-target-actor-split.tsv.gz").open('rb'),
                       compression='gzip',
                       sep="\t",
                       low_memory=False,
                       usecols=['ecco_id', 'actor_id',
                                'actor_role', 'name_unified'],
                       dtype='string')


def _load_multi_target_major_actor_split() -> pd.DataFrame:
    return pd.read_csv(here("data/work/ecco-multi-target-major-actor-split.tsv.gz").open('rb'),
                       compression='gzip',
                       sep="\t",
                       low_memory=False,
                       usecols=['ecco_id', 'major_actor_id',
                                'actor_role', 'name_unified'],
                       dtype='string')


def _split_ids(split_df: pd.DataFrame) -> tuple[list, list, list]:
    d = split_df.reset_index().groupby('set').agg(
        dict(ecco_id=list))['ecco_id'].to_dict()
    return d['train'], d['val'], d['test']


def _create_multi_target_data():
    multi_target_data = _load_multi_target_split()
    train_ids, val_ids, test_ids = _split_ids(
        multi_target_data)
    class_to_class_id: dict[str, dict[str, int]] = dict()
    class_id_to_class: dict[str, list[str]] = dict()
    class_weights: dict[str, torch.Tensor] = dict()
    targetmap: dict[str, torch.Tensor] = dict()
    ids = multi_target_data.reset_index()['ecco_id'].to_list()
    idmap: dict[str, int] = {id: index for index, id in enumerate(ids)}
    for target in ['ecco_part', 'ecco_module', 'primary_language', 'publication_place', 'publication_country', 'major_primary_language', 'major_publication_place', 'major_publication_country']:
        class_pagecounts = {k: v for k,  # type: ignore
                                         v in multi_target_data.groupby(target)['pages'].sum().items()}
        class_to_class_id[target] = dict(  # type: ignore
            map(reversed, enumerate(class_pagecounts.keys())))  # type: ignore
        class_weights[target] = normalised_sqrt_weighted_weights(
            class_pagecounts.values())
        class_id_to_class[target] = [
            k for k in class_to_class_id[target].keys()]
        targetmap[target] = torch.full(
            (len(idmap),), -100, dtype=torch.int8 if len(class_id_to_class[target]) <= 127 else torch.int16)
        for k, v in multi_target_data[target].dropna().items():
            targetmap[target][idmap[cast(str, k)]
                              ] = class_to_class_id[target][v]
    for target in ['total_price', 'earliest_publication_year', 'publication_year', 'pages']:
        targetmap[target] = torch.full(
            (len(idmap),), -100, dtype=torch.float32 if target == 'total_price' else torch.int16)
        for k, v in multi_target_data[target].dropna().items():
            targetmap[target][idmap[cast(str, k)]] = v
    multi_target_major_actor_data = _load_multi_target_major_actor_split()
    class_id_to_class['major_actors'] = multi_target_major_actor_data['major_actor_id'].drop_duplicates(
    ).to_list()
    actor_id_to_actor_id_num = {k: v for v,
                                k in enumerate(class_id_to_class['major_actors'])}
    major_actors = defaultdict(list)
    for ecco_id, actor_id in multi_target_major_actor_data[['ecco_id', 'major_actor_id']].drop_duplicates().itertuples(index=False):
        major_actors[ecco_id].append(
            actor_id_to_actor_id_num[actor_id])
    tensors = [torch.tensor((), dtype=torch.int8) for _ in range(len(idmap))]
    for k, v in major_actors.items():
        tensors[idmap[k]] = torch.tensor(v, dtype=torch.int8)
    targetmap['major_actors'] = torch.nested.nested_tensor(
        tensors, dtype=torch.int8)
    class_pagecounts = {k: v for k, v in multi_target_data.join(
        multi_target_major_actor_data.set_index('ecco_id'), how="inner").groupby('major_actor_id')['pages'].sum().items()}
    class_weights['major_actors'] = normalised_sqrt_weighted_weights(
        [class_pagecounts[k] for k in actor_id_to_actor_id_num.keys()])
    multi_target_actor_data = _load_multi_target_actor_split()
    class_id_to_class['actors'] = class_id_to_class['major_actors'].copy()
    for actor_id in multi_target_actor_data['actor_id']:
        if actor_id not in actor_id_to_actor_id_num:
            actor_id_to_actor_id_num[actor_id] = len(
                class_id_to_class['actors'])
            class_id_to_class['actors'].append(actor_id)
    actors = defaultdict(list)
    for ecco_id, actor_id in multi_target_actor_data[['ecco_id', 'actor_id']].drop_duplicates().itertuples(index=False):
        actors[ecco_id].append(
            actor_id_to_actor_id_num[actor_id])
    tensors = [torch.tensor((), dtype=torch.int32) for _ in range(len(idmap))]
    for k, v in actors.items():
        tensors[idmap[k]] = torch.tensor(v, dtype=torch.int32)
    targetmap['actors'] = torch.nested.nested_tensor(
        tensors, dtype=torch.int32)
    class_pagecounts = {k: v for k, v in multi_target_data.join(
        multi_target_actor_data.set_index('ecco_id'), how="inner").groupby('actor_id')['pages'].sum().items()}
    class_weights['actors'] = normalised_sqrt_weighted_weights(
        [class_pagecounts[k] for k in actor_id_to_actor_id_num.keys()])
    with open(here("hpc_hd_data/data/multi_target_data.pkl"), "wb") as f:
        pickle.dump((ids, torch.tensor([idmap[id] for id in train_ids], dtype=torch.int32), torch.tensor([idmap[id] for id in val_ids], dtype=torch.int32), torch.tensor([idmap[id] for id in test_ids], dtype=torch.int32), class_id_to_class,
                     class_weights, targetmap), f)


def _load_multi_target_data() -> tuple[dict[str, int], list[str], list[str], list[str], dict[str, list[str]], dict[str, torch.Tensor], dict[str, torch.Tensor]]:
    data = resources.read_binary(hpc_hd_data.data, "multi_target_data.pkl")
    ids, train_ids, val_ids, test_ids, class_id_to_class, class_weights, targetmap = pickle.loads(
        data)
    idmap: dict[str, int] = {id: index for index, id in enumerate(ids)}
    return (
        idmap,
        [ids[id] for id in train_ids],
        [ids[id] for id in test_ids],
        [ids[id] for id in val_ids],
        class_id_to_class,
        class_weights,
        targetmap
    )


def nop(*args, **kwargs) -> None:
    pass


def normalised_sqrt_weighted_weights(counts: Iterable[int]) -> torch.Tensor:
    cw = [
        1 / math.sqrt(v) if v != 0 else 0.0 for v in counts]
    tw = sum(cw)
    return torch.tensor([v / tw for v in cw], dtype=torch.float32)


class AECCOMultiTargetDataModule(ABaseDataModule[T_co, T2_co], Generic[Unpack[Ts], T_co, T2_co], abc.ABC):
    def __init__(self,
                 columns_to_return: str = "id, img_offset, img_compressed_size + 128, txt_offset, txt_compressed_size + 128",
                 sqlite_url: str = "s3://ecco-data/ecco_good_page_image_and_text_offsets.sqlite3",
                 image_zip_url: str = "s3://ecco-data/ecco_page_images.zip",
                 text_zip_url: str = "s3://ecco-data/ecco_page_texts.zip",
                 storage_options_yaml_file: str = str(here("s3_secret.yaml")),
                 cache_dir: str = str(here("data/work/cache")),
                 disable_caching: bool = False,
                 train_transform: Callable[[Dataset[tuple[Unpack[Ts]]]],
                                           Dataset[T_co]] = identity_transformation,
                 test_transform: Callable[[Dataset[tuple[Unpack[Ts]]]],
                                          Dataset[T_co]] = identity_transformation,
                 before_setup_hook: Callable[[
                     'AECCOMultiTargetDataModule', Literal['fit', 'validate', 'test', 'predict']], None] = nop,
                 after_setup_hook: Callable[[
                     'AECCOMultiTargetDataModule', Literal['fit', 'validate', 'test', 'predict']], None] = nop,
                 before_prepare_data_hook: Callable[[
                     'AECCOMultiTargetDataModule'], None] = nop,
                 after_prepare_data_hook: Callable[[
                     'AECCOMultiTargetDataModule'], None] = nop,
                 before_teardown_hook: Callable[[
                     'AECCOMultiTargetDataModule'], None] = nop,
                 after_teardown_hook: Callable[[
                     'AECCOMultiTargetDataModule'], None] = nop,
                 **kwargs):
        super().__init__(**kwargs)
        self.storage_options = get_s3fs_credentials(storage_options_yaml_file)
        self.sqlite_url = sqlite_url
        self.image_zip_url = image_zip_url
        self.image_zip_size = get_async_filesystem(
            image_zip_url, self.storage_options).size(image_zip_url)
        self.image_afetcher = PLocalAFetcher(
            image_zip_url, self.image_zip_size, storage_options=self.storage_options, cache_dir=cache_dir if not disable_caching else None)
        self.text_zip_url = text_zip_url
        self.text_zip_size = get_async_filesystem(
            text_zip_url, self.storage_options).size(text_zip_url)
        self.text_afetcher = PLocalAFetcher(
            text_zip_url, self.text_zip_size, storage_options=self.storage_options, cache_dir=cache_dir if not disable_caching else None)
        self.cache_dir = cache_dir
        self.prepare_data_per_node = True
        self.disable_caching = disable_caching
        self.train_transform = train_transform
        self.test_transform = test_transform
        self.before_setup_hook = before_setup_hook
        self.after_setup_hook = after_setup_hook
        self.before_prepare_data_hook = before_prepare_data_hook
        self.after_prepare_data_hook = after_prepare_data_hook
        self.before_teardown_hook = before_teardown_hook
        self.after_teardown_hook = after_teardown_hook
        self.columns_to_return = columns_to_return
        self.sqlite_dataset: SQLiteDataset[tuple[Unpack[Ts]]] = SQLiteDataset(
            sqlite_filename=cache_locally_if_remote(
                self.sqlite_url, storage_options=self.storage_options, cache_dir=self.cache_dir),
            table_name="ecco_pages",
            index_column="entry_number",
            columns_to_return=self.columns_to_return,
            id_column="id")
        self.idmap, self.train_ids, self.val_ids, self.test_ids, self.class_id_to_class, self.class_weights, self.targetmap = _load_multi_target_data()

    def prepare_data(self):
        # download, IO, etc. Useful with shared filesystems
        # only called on 1 GPU/TPU in distributed
        # Ensure sqlite database is downloaded
        self.before_prepare_data_hook(self)
        super().prepare_data()
        cache_locally_if_remote(
            self.sqlite_url, storage_options=self.storage_options, cache_dir=self.cache_dir)
        if not self.disable_caching:
            prefetch_if_remote(self.image_zip_url, self.image_zip_size,
                               self.cache_dir, self.storage_options)
            prefetch_if_remote(self.text_zip_url, self.text_zip_size,
                               self.cache_dir, self.storage_options)
        self.after_prepare_data_hook(self)

    def setup(self, stage: Literal['fit', 'validate', 'test', 'predict']):
        # Called at the beginning of fit (train + validate), validate, test, or predict.
        # This is a good hook when you need to build models dynamically or adjust something
        # about them. This hook is called on every process when using DDP.
        self.before_setup_hook(self, stage)
        super().setup(stage)
        # if (stage == "fit") and self.train_dataset is None:
        self.train_dataset = self.train_transform(
            ECCOIdBasedSubdataset(self.train_ids, self.sqlite_dataset))
        # if (stage == "fit" or stage == "validate") and self.val_dataset is None:
        self.val_dataset = self.test_transform(
            ECCOIdBasedSubdataset(self.val_ids, self.sqlite_dataset))
        # if (stage == "test") and self.test_dataset is None:
        self.test_dataset = self.test_transform(
            ECCOIdBasedSubdataset(self.test_ids, self.sqlite_dataset))
        self.after_setup_hook(self, stage)

    def teardown(self, stage: Literal['fit', 'validate', 'test', 'predict']):
        # clean up state after the trainer stops, delete files...
        # called on every process in DDP
        self.before_teardown_hook(self)
        super().teardown(stage)
        self.after_teardown_hook(self)


def _contrastive_ecco_images_transform(offsets_sizes_and_ids: Sequence[tuple[str, int, int, Unpack[Ts]]], afetcher: AFetcher, image_transform: Callable[[Image.Image], T_co]) -> Sequence[tuple[str, T_co, T_co]]:
    images: list[Image.Image | None] = _load_from_zip_uncompress_and_transform(
        [(offset, size) for _, offset, size, *_ in offsets_sizes_and_ids],
        afetcher=afetcher, transform=functools.partial(_load_and_transform_image, transform=identity_transformation))
    # TODO: more than 2 views
    return _filter_out_nones(
        zip(
            [id for id, *_ in offsets_sizes_and_ids],
            map(image_transform, images),  # type: ignore
            map(image_transform, images)  # type: ignore
        )
    )


def _contrastive_ecco_images_ds_transform(ds: Dataset[tuple[str, int, int, Unpack[Ts]]], afetcher: AFetcher, image_transform: Callable[[Image.Image], T_co]) -> Dataset[tuple[str, T_co, T_co]]:
    return EntryTransformingMapDataset(dataset=ds, transform=functools.partial(_contrastive_ecco_images_transform, afetcher=afetcher, image_transform=image_transform))


class ContrastiveECCOImagesDataModule(AECCOMultiTargetDataModule[str, int, int, tuple[str, torch.Tensor, torch.Tensor], tuple[tuple[str], torch.Tensor, torch.Tensor]]):

    def __init__(self,
                 image_test_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 image_train_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 **kwargs):
        super().__init__(
            columns_to_return="id, img_offset, img_compressed_size + 128",
            train_transform=functools.partial(_contrastive_ecco_images_ds_transform, afetcher=self.image_afetcher,
                                              image_transform=image_train_transform),
            test_transform=functools.partial(_contrastive_ecco_images_ds_transform, afetcher=self.image_afetcher,
                                             image_transform=image_test_transform),
            **kwargs
        )
        self.save_hyperparameters(
            ignore=['train_transform', 'test_transform'])


def _to_string_and_transform(d: BytesIO, transform: Callable[[str], T_co]) -> T_co:
    return transform(d.getvalue().decode("utf-8"))


class MultiModalItem(TypedDict, Generic[T_co], total=False):
    id: str
    image: torch.Tensor | None
    text: T_co | None
    ecco_part: int
    ecco_module: int
    publication_year: int
    earliest_publication_year: int
    primary_language: int
    total_price: float
    publication_place: int
    publication_country: int
    pages: int
    major_primary_language: int
    major_publication_place: int
    major_publication_country: int
    actors: torch.Tensor
    major_actors: torch.Tensor


def multimodal_item(id: str, idmap: dict[str, int], targetmap: dict[str, torch.Tensor], targets: list[str], n_actors: int, n_major_actors: int) -> MultiModalItem[Any]:
    ret = MultiModalItem()
    sid = id[:10]
    for target in targets:
        if target == 'id':
            ret['id'] = id
        elif target == 'actors':
            t_a = targetmap[target][idmap[sid]]
            tensor = torch.full(
                size=(n_actors,), fill_value=-1, dtype=torch.int64)
            tensor[0:len(t_a)] = t_a
            ret[target] = tensor
        elif target == 'major_actors':
            t_a = targetmap[target][idmap[sid]]
            tensor = torch.full(
                size=(n_major_actors,), fill_value=-1, dtype=torch.int64)
            tensor[0:len(t_a)] = t_a
            ret[target] = tensor
        else:
            ret[target] = targetmap[target][idmap[sid]]
    return ret


def _ecco_images_and_texts_transform(ids_offsets_and_sizes: Sequence[tuple[str, int, int, int, int]], image_afetcher: AFetcher, image_transform: Callable[[Image.Image], torch.Tensor], text_afetcher: AFetcher, text_transform: Callable[[str], T_co], idmap: dict[str, int], targetmap: dict[str, torch.Tensor], targets: list[str], features: list[str], n_actors: int, n_major_actors: int) -> Sequence[MultiModalItem[T_co]]:
    ret = [multimodal_item(id, idmap, targetmap, targets, n_actors, n_major_actors)
           for id, *_ in ids_offsets_and_sizes]
    if 'text' in features:
        texts = _load_from_zip_uncompress_and_transform(
            [(offset, size)
             for _, _, _, offset, size in ids_offsets_and_sizes],
            afetcher=text_afetcher,
            transform=functools.partial(
                _to_string_and_transform, transform=text_transform)  # type: ignore
        )
        for item, text in zip(ret, texts):
            item['text'] = text
    if 'image' in features:
        images = _load_from_zip_uncompress_and_transform(
            [(offset, size)
             for _, offset, size, _, _ in ids_offsets_and_sizes],
            afetcher=image_afetcher,
            transform=functools.partial(
                _load_and_transform_image, transform=image_transform)
        )
        for item, image in zip(ret, images):
            item['image'] = image
    return ret


def _ecco_images_and_texts_to_targets_ds_transform(ds: Dataset[tuple[str, int, int, int, int]], image_afetcher: AFetcher, image_transform: Callable[[Image.Image], torch.Tensor], text_afetcher: AFetcher, text_transform: Callable[[str], T_co], idmap: dict[str, int], targetmap: dict[str, torch.Tensor], targets: list[str], features: list[str], n_actors: int, n_major_actors: int) -> Dataset[MultiModalItem[T_co]]:
    return EntryTransformingMapDataset(dataset=ds, transform=functools.partial(_ecco_images_and_texts_transform, image_afetcher=image_afetcher, image_transform=image_transform, text_afetcher=text_afetcher, text_transform=text_transform, idmap=idmap, targetmap=targetmap, targets=targets, features=features, n_actors=n_actors, n_major_actors=n_major_actors))


class MultiModalBatch(TypedDict, Generic[T_co], total=False):
    id: Sequence[str]
    image: torch.Tensor
    text: T_co
    ecco_part: torch.Tensor
    ecco_module: torch.Tensor
    publication_year: torch.Tensor
    earliest_publication_year: torch.Tensor
    primary_language: torch.Tensor
    total_price: torch.Tensor
    publication_place: torch.Tensor
    publication_country: torch.Tensor
    pages: torch.Tensor
    major_primary_language: torch.Tensor
    major_publication_place: torch.Tensor
    major_publication_country: torch.Tensor
    actors: torch.Tensor
    major_actors: torch.Tensor


def _tokenise_batch(text: Sequence[str], tokenizer, max_length: int = 512, padding: str = 'max_length', truncation: bool = True) -> BatchEncoding:
    with torch.no_grad():
        return tokenizer(list(text), max_length=max_length, padding=padding, truncation=truncation, return_tensors="pt")


def _embed_batch(text: Sequence[str], tokenizer, model) -> torch.Tensor:
    with torch.no_grad():
        inputs = _tokenise_batch(text, tokenizer)
        outputs = model(**inputs)
    return outputs.last_hidden_state.squeeze(0)


def batch_embed_after_collate(batch: Sequence[MultiModalItem[str]], tokenizer, model, collate_fn: Callable[[Sequence[MultiModalItem[str]]], MultiModalBatch[str]]) -> MultiModalBatch[torch.Tensor]:
    cbatch = collate_fn(batch)
    if 'text' in cbatch:
        cbatch['text'] = _embed_batch(  # type: ignore
            cbatch['text'], tokenizer, model)  # type: ignore
    return cbatch  # type: ignore


def batch_tokenise_after_collate(batch: Sequence[Any], tokenizer, collate_fn: Callable[[Sequence[MultiModalItem[str]]], MultiModalBatch[str]]) -> MultiModalBatch[BatchEncoding]:
    cbatch = collate_fn(batch)
    if 'text' in cbatch:
        cbatch['text'] = _tokenise_batch(  # type: ignore
            cbatch['text'], tokenizer)  # type: ignore
    return cbatch  # type: ignore


def filter_out_bad_entries(batch: Sequence[MultiModalItem[T_co]], collate_fn: Callable[[Sequence[MultiModalItem[T_co]]], MultiModalBatch[T2_co]] = cast(Callable[[Sequence[MultiModalItem[T_co]]], MultiModalBatch[T2_co]], torch.utils.data.dataloader.default_collate)) -> MultiModalBatch[T2_co]:
    return collate_fn([i for i in batch if ('text' not in i or i['text'] is not None) and ('image' not in i or i['image'] is not None)])


def setup_batch_tokenise_after_collate(dm: AECCOMultiTargetDataModule, _: Literal['fit', 'validate', 'test', 'predict']):
    dm.collate_fn = functools.partial(batch_tokenise_after_collate, tokenizer=AutoTokenizer.from_pretrained(
        "TurkuNLP/eccobert-base-cased-v1"), collate_fn=cast(Callable[[Sequence[MultiModalItem[str]]], MultiModalBatch[str]], filter_out_bad_entries))


def setup_batch_embed_after_collate(dm: AECCOMultiTargetDataModule, _: Literal['fit', 'validate', 'test', 'predict']):
    dm.collate_fn = functools.partial(batch_embed_after_collate, tokenizer=AutoTokenizer.from_pretrained(
        "TurkuNLP/eccobert-base-cased-v1"), model=AutoModel.from_pretrained("TurkuNLP/eccobert-base-cased-v1"), collate_fn=cast(Callable[[Sequence[MultiModalItem[str]]], MultiModalBatch[str]], filter_out_bad_entries))


def collate_tensors_with_different_dims_as_nested_tensors(batch, *args, **kwargs):
    it = iter(batch)

    elem_shape = next(it).shape
    if not all(elem.shape == elem_shape for elem in it):
        return torch.nested.nested_tensor(batch)
    else:
        return torch.utils.data._utils.collate.collate_tensor_fn(batch, *args, **kwargs)  # type: ignore # noqa


# torch.utils.data._utils.collate.default_collate_fn_map.update(  # type: ignore # noqa
#    {torch.Tensor: collate_tensors_with_different_dims_as_nested_tensors})


all_targets = ['id', 'ecco_part', 'ecco_module', 'publication_year', 'earliest_publication_year', 'primary_language', 'total_price',
               'publication_place', 'publication_country', 'pages', 'major_primary_language', 'major_publication_place', 'major_publication_country', 'actors', 'major_actors']


def map_target_type(target: str) -> Literal['single-label', 'multi-label', 'regression']:
    if target in {'publication_year', 'earliest_publication_year', 'total_price', 'pages'}:
        return 'regression'
    elif target in {'actors', 'major_actors'}:
        return 'multi-label'
    return 'single-label'


class ECCOImagesAndTextsToTargetsDataModule(AECCOMultiTargetDataModule[str, int, int, int, int, MultiModalItem[T_co], MultiModalBatch[T2_co]], Generic[T_co, T2_co]):
    def __init__(self,
                 image_train_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 image_test_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 text_train_transform: Callable[[
                     str], T_co] = identity_transformation,
                 text_test_transform: Callable[[
                     str], T_co] = identity_transformation,
                 collate_fn: Callable[[Sequence[MultiModalItem[T_co]]],
                                      MultiModalBatch[T2_co]] = cast(Callable[[Sequence[MultiModalItem[T_co]]], MultiModalBatch[T2_co]], filter_out_bad_entries),
                 before_setup_hook: Callable[[
                     AECCOMultiTargetDataModule, Literal['fit', 'validate', 'test', 'predict']], None] = setup_batch_tokenise_after_collate,
                 targets: list[str] = all_targets,
                 features: list[str] = ['image', 'text'],
                 **kwargs):
        super().__init__(
            columns_to_return="id, img_offset, img_compressed_size + 128, txt_offset, txt_compressed_size + 128",
            collate_fn=collate_fn,
            before_setup_hook=before_setup_hook,
            **kwargs)
        self.targets = targets
        self.features = features
        self.train_transform = functools.partial(
            _ecco_images_and_texts_to_targets_ds_transform, image_afetcher=self.image_afetcher, image_transform=image_train_transform, text_afetcher=self.text_afetcher, text_transform=text_train_transform, idmap=self.idmap, targetmap=self.targetmap, targets=targets, features=features, n_actors=len(self.class_id_to_class['actors']), n_major_actors=len(self.class_id_to_class['major_actors']))
        self.test_transform = functools.partial(
            _ecco_images_and_texts_to_targets_ds_transform, image_afetcher=self.image_afetcher, image_transform=image_test_transform, text_afetcher=self.text_afetcher, text_transform=text_test_transform, idmap=self.idmap, targetmap=self.targetmap, targets=targets, features=features, n_actors=len(self.class_id_to_class['actors']), n_major_actors=len(self.class_id_to_class['major_actors']))
        self.save_hyperparameters()
