from transformers import AutoTokenizer, AutoModel, BatchEncoding
import abc
import functools
from io import BytesIO
from typing import Any, Callable, Generic, Iterable, Sequence, TypeVar, TypeVarTuple
import fsspec
from hereutil import here
import pandas
import torch
import torch.utils.data.dataloader
from PIL import Image
from datar import f
import datar.base as d
from datar.dplyr import pull, group_by, summarise, mutate
from datar.tibble import as_tibble
import numpy as np
from torch.utils.data import Dataset
from stream_unzip import stream_unzip
from hscitorchutil.dataset import TransformedMapDataset, UnionMapDataset
from hscitorchutil.sqlite import SQLiteDataModule, SQLiteDataset, _remove_nones_from_batch
from hscitorchutil.fsspec import fetch_async, AFetcher, get_async_filesystem, get_s3fs_credentials, prefetch_if_remote, PLocalAFetcher

import logging

T_co = TypeVar('T_co', covariant=True)
T2_co = TypeVar('T2_co', covariant=True)
Ts = TypeVarTuple("Ts")


class AECCOImagesDataModule(SQLiteDataModule[int, int, str, T_co, T2_co], Generic[T_co,T2_co], abc.ABC):
    def __init__(self,
                 train_sqlite_url: str = "s3://ecco-data/ecco_pages_train.sqlite3",
                 val_sqlite_url: str = "s3://ecco-data/ecco_pages_val.sqlite3",
                 test_sqlite_url: str = "s3://ecco-data/ecco_pages_test.sqlite3",
                 zip_url: str = "s3://ecco-data/ecco_page_images.zip",
                 table_name: str = "ecco_pages",
                 index_column: str = "entry_number",
                 columns_to_return: str = "img_offset, img_compressed_size + 128, id",
                 id_column: str = "id",
                 cache_dir: str = str(here("data/work/ecco_modules/cache")),
                 storage_options_yaml_file: str = str(here("s3_secret.yaml")),
                 disable_caching: bool = False,
                 **kwargs):
        self.storage_options = get_s3fs_credentials(storage_options_yaml_file)
        self.disable_caching = disable_caching
        self.zip_url = zip_url
        self.zip_size = get_async_filesystem(
            zip_url, self.storage_options).size(zip_url)
        self.afetcher = PLocalAFetcher(
            zip_url, self.zip_size, storage_options=self.storage_options, cache_dir=cache_dir if not disable_caching else None)
        super().__init__(
            train_sqlite_url=train_sqlite_url,
            val_sqlite_url=val_sqlite_url,
            test_sqlite_url=test_sqlite_url,
            storage_options=self.storage_options,
            cache_dir=cache_dir,
            table_name=table_name,
            index_column=index_column,
            columns_to_return=columns_to_return,
            id_column=id_column,
            **kwargs)

    def prepare_data(self):
        # download, IO, etc. Useful with shared filesystems
        # only called on 1 GPU/TPU in distributed

        # Ensure sqlite databases are downloaded
        super().prepare_data()
        # Start a background process to read the ZIP file into our mmap cache if zip_url is in S3
        if not self.disable_caching:
            prefetch_if_remote(self.zip_url, self.zip_size,
                               self.cache_dir, self.storage_options)


def transform_binary_PIL_image_to_bool_tensor(img: Image.Image) -> torch.Tensor:
    """Convert a PIL Image to a torch bool tensor"""
    return (
        torch.from_numpy(np.array(img, dtype=bool))
        .view(img.size[1], img.size[0], 1)
        .permute((2, 0, 1))
        .contiguous()  # HWC to CHW
    )


def identity_transformation(x: T_co) -> T_co:
    return x


def _load_and_transform_image(image_bytes: BytesIO, transform: Callable[[Image.Image], T_co]) -> T_co | None:
    try:
        im = Image.open(image_bytes)
        return transform(im)
    except:
        logging.exception("Failed to load/transform image, returning None")
        return None


def _uncompress(compressed_bytes: bytes) -> BytesIO:
    _, _, biter = next(stream_unzip([compressed_bytes]))
    bio = BytesIO()
    for b in biter:
        bio.write(b)
    return bio


def _load_from_zip_uncompress_and_transform(offsets_and_sizes: Sequence[tuple[int, int]], afetcher: AFetcher, transform: Callable[[BytesIO], T_co]) -> list[T_co]:
    try:
        return [transform(_uncompress(compressed_bytes)) for compressed_bytes in fetch_async(afetcher, [(offset, offset+size) for offset, size in offsets_and_sizes])]
    except Exception as e:
        logging.exception("Failed to load texts, returning empty batch")
        return []

def _filter_out_nones(tuples: Iterable[tuple[*Ts]]) -> Sequence[tuple[*Ts]]:
    return list(filter(lambda x: not any(map(lambda x: x is None, x)), tuples))


def _ecco_images_transform(offsets_sizes_and_ids: Sequence[tuple[int, int, str]], afetcher: AFetcher, image_transform: Callable[[Image.Image], T_co]) -> Sequence[tuple[str, T_co]]:
    return _filter_out_nones(
        zip(
            [id for _, _, id in offsets_sizes_and_ids],
            _load_from_zip_uncompress_and_transform(
                [(offset, size)
                 for offset, size, _ in offsets_sizes_and_ids],
                afetcher=afetcher,
                transform=functools.partial(
                    _load_and_transform_image, transform=image_transform)  # type: ignore
            )
        )
    )


def _get_image_from_tuple(t: tuple[str, torch.Tensor]) -> torch.Tensor:
    return t[1]

class ECCOImagesDataModule(AECCOImagesDataModule[tuple[str, torch.Tensor] | None, tuple[tuple[str], torch.Tensor] | None]):

    def __init__(self,
                 image_test_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 image_train_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 **kwargs):
        super().__init__(**kwargs)
        self.train_transform = functools.partial(
            _ecco_images_transform, afetcher=self.afetcher, image_transform=image_train_transform)
        self.test_transform = functools.partial(
            _ecco_images_transform, afetcher=self.afetcher, image_transform=image_test_transform)
        self.save_hyperparameters(
            ignore=['image_train_transforms', 'image_test_transforms'])


def _contrastive_ecco_images_transform(offsets_sizes_and_ids: Sequence[tuple[int, int, str]], afetcher: AFetcher, image_transform: Callable[[Image.Image], T_co]) -> Sequence[tuple[str, T_co, T_co]]:
    images: list[Image.Image | None] = _load_from_zip_uncompress_and_transform(
        [(offset, size) for offset, size, _ in offsets_sizes_and_ids],
        afetcher=afetcher, transform=functools.partial(_load_and_transform_image, transform=identity_transformation))
    # TODO: more than 2 views
    return _filter_out_nones(
        zip(
            [id for _, _, id in offsets_sizes_and_ids],
            map(image_transform, images),  # type: ignore
            map(image_transform, images)  # type: ignore
        )
    )


class ContrastiveECCOImagesDataModule(AECCOImagesDataModule[tuple[str, torch.Tensor, torch.Tensor]]):

    def __init__(self,
                 image_test_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 image_train_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 **kwargs):
        super().__init__(**kwargs)
        self.train_transform = functools.partial(_contrastive_ecco_images_transform, afetcher=self.afetcher,
                                                 image_transform=image_train_transform)
        self.test_transform = functools.partial(_contrastive_ecco_images_transform, afetcher=self.afetcher,
                                                image_transform=image_test_transform)
        self.save_hyperparameters(
            ignore=['transforms'])


def _ecco_images_to_ecco_modules_transform(offsets_sizes_and_ids: Sequence[tuple[int, int, str]], afetcher: AFetcher, image_transform: Callable[[Image.Image], T_co], ecco_id_to_class_id: dict[str, int], unknown_id: int) -> list[tuple[str, T_co, int]]:
    return _filter_out_nones(
        zip(
            [id[:10] for _, _, id in offsets_sizes_and_ids],
            _load_from_zip_uncompress_and_transform(
                [(offset, size)
                 for offset, size, _ in offsets_sizes_and_ids],
                afetcher=afetcher,
                transform=functools.partial(
                    _load_and_transform_image, transform=image_transform)  # type: ignore
            ),
            [ecco_id_to_class_id.get(id[:10], unknown_id) for _, _, id in offsets_sizes_and_ids]))


def _get_images_from_test_dataset(train_dataset: Dataset[tuple[str, torch.Tensor]], val_dataset: Dataset[tuple[str, torch.Tensor]], test_dataset: Dataset[tuple[str, torch.Tensor]]) -> Dataset[torch.Tensor]:
    return TransformedMapDataset(test_dataset, _get_image_from_tuple)

class ECCOModuleTargetDataModule(TransformedMapDataset[tuple[str, *Ts]],Generic[*Ts]):
    def __init__(self, 
                 feature_dataset: Dataset[tuple[str, *Ts]],
                 storage_options_yaml_file: str = str(here("s3_secret.yaml")),
                 ecco_core_parquet_location: str = "s3://ecco-data/ecco_core.parquet"):
        with fsspec.open(ecco_core_parquet_location, **get_s3fs_credentials(storage_options_yaml_file)) as pf:
            self.ecco_core = as_tibble(pandas.read_parquet(pf))  # type: ignore
        classes_and_weights = (self.ecco_core >>
                    group_by(f.ecco_module) >>
                    summarise(pages=d.sum(f.ecco_pages)) >>  # type: ignore # noqa
                    mutate(weight=1/d.sqrt(f.pages)) >>  # type: ignore # noqa
                    mutate(weight=f.weight/d.sum(f.weight))  # type: ignore # noqa
                    )
        self.classes: list[str] = classes_and_weights >> pull(
            f.ecco_module, to='list')
        self.classes = self.classes + ["unknown"]
        self.class_to_class_id: dict[str, int] = dict(
            map(reversed, enumerate(self.classes)))  # type: ignore
        self.unknown_id = len(self.classes)-1
        ecco_id_to_class_id = dict(zip(self.ecco_core.ecco_id, map(
            lambda ecco_module: self.class_to_class_id[ecco_module], self.ecco_core.ecco_module)))
        # Calculate class weights for use in loss function using sqrt inverse frequency weighting.
        self.class_weights = [
            *(classes_and_weights >> pull(f.weight, to='list')), 0.0]
        self.num_classes = len(self.classes)        
        super().__init__(dataset=feature_dataset, transform=functools.partial)

class ECCOImagesToECCOModulesDataModule(AECCOImagesDataModule[tuple[str, torch.Tensor, int]]):
    def __init__(self,
                 image_test_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 image_train_transform: Callable[[
                     Image.Image], torch.Tensor] = transform_binary_PIL_image_to_bool_tensor,
                 storage_options_yaml_file: str = str(here("s3_secret.yaml")),
                 ecco_core_parquet_location: str = "s3://ecco-data/ecco_core.parquet",
                 get_predict_dataset: Callable[[Dataset[tuple[str, torch.Tensor]], Dataset[tuple[str, torch.Tensor]],
                                                Dataset[tuple[str, torch.Tensor]]], Dataset[torch.Tensor]] = _get_images_from_test_dataset,
                 **kwargs):
        super().__init__(
            storage_options_yaml_file=storage_options_yaml_file,
            columns_to_return="img_offset, img_compressed_size + 128, id",
            get_predict_dataset=get_predict_dataset,
            **kwargs)
        with fsspec.open(ecco_core_parquet_location, **get_s3fs_credentials(storage_options_yaml_file)) as pf:
            self.ecco_core = as_tibble(pandas.read_parquet(pf))  # type: ignore
        classes_and_weights = (self.ecco_core >>
                    group_by(f.ecco_module) >>
                    summarise(pages=d.sum(f.ecco_pages)) >>  # type: ignore # noqa
                    mutate(weight=1/d.sqrt(f.pages)) >>  # type: ignore # noqa
                    mutate(weight=f.weight/d.sum(f.weight))  # type: ignore # noqa
                    )
        self.classes: list[str] = classes_and_weights >> pull(
            f.ecco_module, to='list')
        self.classes = self.classes + ["unknown"]
        self.class_to_class_id: dict[str, int] = dict(
            map(reversed, enumerate(self.classes)))  # type: ignore
        self.unknown_id = len(self.classes)-1
        ecco_id_to_class_id = dict(zip(self.ecco_core.ecco_id, map(
            lambda ecco_module: self.class_to_class_id[ecco_module], self.ecco_core.ecco_module)))
        # Calculate class weights for use in loss function using sqrt inverse frequency weighting.
        self.class_weights = [
            *(classes_and_weights >> pull(f.weight, to='list')), 0.0]
        self.num_classes = len(self.classes)
        self.train_transform = functools.partial(_ecco_images_to_ecco_modules_transform, afetcher=self.afetcher,
                                                 image_transform=image_train_transform, ecco_id_to_class_id=ecco_id_to_class_id, unknown_id=self.unknown_id)
        self.test_transform = functools.partial(_ecco_images_to_ecco_modules_transform, afetcher=self.afetcher,
                                                image_transform=image_test_transform, ecco_id_to_class_id=ecco_id_to_class_id, unknown_id=self.unknown_id)
        self.save_hyperparameters(
            ignore=['image_train_transforms', 'image_test_transforms'])


def _to_string_and_transform(d: BytesIO, transform: Callable[[str], T_co]) -> T_co:
    return transform(d.getvalue().decode("utf-8"))


def _ecco_page_texts_transform(offsets_sizes_and_ids: Sequence[tuple[int, int, str]], afetcher: AFetcher, transform: Callable[[str], T_co]) -> Sequence[tuple[str, T_co]]:
    return _filter_out_nones(zip(
        [id for _, _, id in offsets_sizes_and_ids],
        _load_from_zip_uncompress_and_transform([(offset, size) for offset, size, _ in offsets_sizes_and_ids],
                                                afetcher=afetcher, transform=functools.partial(_to_string_and_transform, transform=transform))
    ))


class AECCOPageTextsDataModule(SQLiteDataModule[tuple[int, int, str], T_co]):
    def __init__(self,
                 train_sqlite_url: str = "s3://ecco-data/ecco_pages_train.sqlite3",
                 val_sqlite_url: str = "s3://ecco-data/ecco_pages_val.sqlite3",
                 test_sqlite_url: str = "s3://ecco-data/ecco_pages_test.sqlite3",
                 zip_url: str = "s3://ecco-data/ecco_page_texts.zip",
                 cache_dir: str = str(here("data/work/ecco_modules/cache")),
                 storage_options_yaml_file: str = str(here("s3_secret.yaml")),
                 disable_caching: bool = False,
                 **kwargs):
        self.storage_options = get_s3fs_credentials(storage_options_yaml_file)
        super().__init__(
            train_sqlite_url=train_sqlite_url,
            val_sqlite_url=val_sqlite_url,
            test_sqlite_url=test_sqlite_url,
            storage_options=self.storage_options,
            table_name="ecco_pages",
            index_column="entry_number",
            columns_to_return="txt_offset, txt_compressed_size + 128, id",
            id_column="id",
            cache_dir=cache_dir,
            **kwargs)
        self.disable_caching = disable_caching
        self.zip_url = zip_url
        self.zip_size = get_async_filesystem(
            zip_url, self.storage_options).size(zip_url)
        self.afetcher = PLocalAFetcher(
            zip_url, self.zip_size, storage_options=self.storage_options, cache_dir=self.cache_dir if not disable_caching else None)

    def prepare_data(self):
        # Ensure sqlite databases are downloaded
        super().prepare_data()
        # Start a background process to read the ZIP file into our mmap cache if zip_url is in S3
        if not self.disable_caching:
            prefetch_if_remote(self.zip_url, self.zip_size,
                               self.cache_dir, self.storage_options)


class ECCOPageTextsDataModule(AECCOPageTextsDataModule[tuple[str, T_co]]):
    def __init__(self,
                 text_train_transform: Callable[[str],
                                                T_co] = identity_transformation,
                 text_test_transform: Callable[[str],
                                               T_co] = identity_transformation,
                 **kwargs):
        super().__init__(**kwargs)
        self.train_transform = functools.partial(
            _ecco_page_texts_transform, afetcher=self.afetcher, transform=text_train_transform)
        self.test_transform = functools.partial(
            _ecco_page_texts_transform, afetcher=self.afetcher, transform=text_test_transform)
        self.save_hyperparameters(
            ignore=['text_train_transforms', 'text_test_transforms'])


def _tokenise_batch(text: Sequence[str], tokenizer, max_length: int = 512, padding: str = 'max_length', truncation: bool = True) -> BatchEncoding:
    with torch.no_grad():
        return tokenizer(list(text), max_length=max_length, padding=padding,
                         truncation=truncation, return_tensors="pt")


def _batch_tokenise_after_collate(batch: Sequence[Any], tokenizer, text_index: int, collate_fn: Callable) -> tuple | None:
    cbatch = collate_fn(batch)
    if cbatch is None:
        return None
    return (*cbatch[:text_index], _tokenise_batch(cbatch[text_index],
                                                  tokenizer), *cbatch[text_index+1:])


class ECCOPageTextTokensDataModule(AECCOPageTextsDataModule[tuple[str, str]]):
    def __init__(self, model: str = "TurkuNLP/eccobert-base-cased-v1", **kwargs):
        self.model = model
        super().__init__(
            collate_fn=None,
            **kwargs)
        self.train_transform = functools.partial(
            _ecco_page_texts_transform, afetcher=self.afetcher, transform=identity_transformation)
        self.test_transform = functools.partial(
            _ecco_page_texts_transform, afetcher=self.afetcher, transform=identity_transformation)

    def prepare_data(self):
        super().prepare_data()
        AutoTokenizer.from_pretrained(self.model)

    def setup(self, stage: str | None):
        if self.collate_fn is None:
            self.collate_fn = functools.partial(_batch_tokenise_after_collate, tokenizer=AutoTokenizer.from_pretrained(
                self.model), text_index=1, collate_fn=_remove_nones_from_batch)
        super().setup(stage)


def _ecco_page_text_tokens_to_ecco_modules_transform(offsets_sizes_and_ids: Sequence[tuple[int, int, str]], afetcher: AFetcher, ecco_id_to_class_id: dict[str, int], unknown_id: int) -> Sequence[tuple[str, str, int]]:
    return _filter_out_nones(zip(
        [id for _, _, id in offsets_sizes_and_ids],
        _load_from_zip_uncompress_and_transform([(offset, size) for offset, size, _ in offsets_sizes_and_ids],
                                                afetcher=afetcher, transform=functools.partial(_to_string_and_transform, transform=identity_transformation)),
        [ecco_id_to_class_id.get(id[:10], unknown_id)
         for _, _, id in offsets_sizes_and_ids]
    ))


class ECCOPageTextTokensToECCOModulesDataModule(AECCOPageTextsDataModule[tuple[str, str, int]]):
    def __init__(self,
                 model: str = "TurkuNLP/eccobert-base-cased-v1",
                 storage_options_yaml_file: str = str(here("s3_secret.yaml")),
                 ecco_core_parquet_location: str = "s3://ecco-data/ecco_core.parquet",
                 **kwargs):
        super().__init__(
            storage_options_yaml_file=storage_options_yaml_file,
            collate_fn=None,
            **kwargs)
        self.model = model
        with fsspec.open(ecco_core_parquet_location, **get_s3fs_credentials(storage_options_yaml_file)) as pf:
            self.ecco_core = as_tibble(pandas.read_parquet(pf))  # type: ignore
        classes_and_weights = (self.ecco_core >>
                    group_by(f.ecco_module) >>
                    summarise(pages=d.sum(f.ecco_pages)) >>  # type: ignore # noqa
                    mutate(weight=1/d.sqrt(f.pages)) >>  # type: ignore # noqa
                    mutate(weight=f.weight/d.sum(f.weight))  # type: ignore # noqa
                    )
        self.classes: list[str] = classes_and_weights >> pull(
            f.ecco_module, to='list')
        self.classes = self.classes + ["unknown"]
        self.class_to_class_id: dict[str, int] = dict(
            map(reversed, enumerate(self.classes)))  # type: ignore
        self.unknown_id = len(self.classes)-1
        self.ecco_id_to_class_id = dict(zip(self.ecco_core.ecco_id, map(
            lambda ecco_module: self.class_to_class_id[ecco_module], self.ecco_core.ecco_module)))
        # Calculate class weights for use in loss function using sqrt inverse frequency weighting.
        self.class_weights = [
            *(classes_and_weights >> pull(f.weight, to='list')), 0.0]
        self.num_classes = len(self.classes)
        self.train_transform = functools.partial(
            _ecco_page_text_tokens_to_ecco_modules_transform, afetcher=self.afetcher, ecco_id_to_class_id=self.ecco_id_to_class_id, unknown_id=self.unknown_id)
        self.test_transform = functools.partial(
            _ecco_page_text_tokens_to_ecco_modules_transform, afetcher=self.afetcher, ecco_id_to_class_id=self.ecco_id_to_class_id, unknown_id=self.unknown_id)
        self.save_hyperparameters(
            ignore=['text_train_transforms', 'text_test_transforms'])

    def prepare_data(self):
        super().prepare_data()
        AutoTokenizer.from_pretrained(self.model)

    def setup(self, stage: str | None):
        if self.collate_fn is None:
            self.collate_fn = functools.partial(_batch_tokenise_after_collate, tokenizer=AutoTokenizer.from_pretrained(
                self.model), text_index=1, collate_fn=_remove_nones_from_batch)
        super().setup(stage)


def _embed_batch(text: Sequence[str], tokenizer, model) -> torch.Tensor:
    with torch.no_grad():
        inputs = _tokenise_batch(text, tokenizer)
        outputs = model(**inputs)
        return outputs.last_hidden_state.squeeze(0)


def _batch_embed_after_collate(batch: Sequence[Any], tokenizer, model, text_index: int, collate_fn: Callable) -> tuple:
    batch = collate_fn(batch)
    return (*batch[:text_index], _embed_batch(batch[text_index], tokenizer, model), *batch[text_index+1:])


class ECCOPageTextEmbeddingsDataModule(AECCOPageTextsDataModule[tuple[str, str]]):
    def __init__(self, model: str = "TurkuNLP/eccobert-base-cased-v1", **kwargs):
        self.model = model
        super().__init__(
            collate_fn=None,
            **kwargs)
        self.train_transform = functools.partial(
            _ecco_page_texts_transform, afetcher=self.afetcher, transform=identity_transformation)
        self.test_transform = functools.partial(
            _ecco_page_texts_transform, afetcher=self.afetcher, transform=identity_transformation)
        self.orig_collate_fn = self.collate_fn

    def prepare_data(self):
        super().prepare_data()
        AutoTokenizer.from_pretrained(self.model)
        AutoModel.from_pretrained(self.model)

    def setup(self, stage: str | None):
        if self.collate_fn is None:
            self.collate_fn = functools.partial(_batch_embed_after_collate, tokenizer=AutoTokenizer.from_pretrained(
                self.model), model=AutoModel.from_pretrained(self.model), text_index=1, collate_fn=_remove_nones_from_batch)
        super().setup(stage)


def _ecco_images_and_page_texts_transform(offsets_sizes_and_ids: Sequence[tuple[int, int, int, int, str]], image_afetcher: AFetcher, text_afetcher: AFetcher, image_transform: Callable[[Image.Image], T_co], text_transform: Callable[[str], T2_co]) -> Sequence[tuple[str, T_co, T2_co]]:
    return _filter_out_nones(zip(
        [id for _, _, _, _, id in offsets_sizes_and_ids],
        _load_from_zip_uncompress_and_transform([(offset, size) for offset, size, _, _, _ in offsets_sizes_and_ids], afetcher=image_afetcher,
                                                transform=functools.partial(_load_and_transform_image, transform=image_transform)),  # type: ignore
        _load_from_zip_uncompress_and_transform([(offset, size) for _, _, offset, size, _ in offsets_sizes_and_ids],
                                                afetcher=text_afetcher, transform=functools.partial(_to_string_and_transform, transform=text_transform))  # type:ignore
    ))


class ECCOImagesAndPageTextsDataModule(SQLiteDataModule[tuple[int, int, int, int, str], tuple[str, T_co, T2_co]]):
    def __init__(self,
                 train_sqlite_url: str = "s3://ecco-data/ecco_pages_train.sqlite3",
                 val_sqlite_url: str = "s3://ecco-data/ecco_pages_val.sqlite3",
                 test_sqlite_url: str = "s3://ecco-data/ecco_pages_test.sqlite3",
                 image_zip_url: str = "s3://ecco-data/ecco_page_images.zip",
                 text_zip_url: str = "s3://ecco-data/ecco_page_texts.zip",
                 cache_dir: str = str(here("data/work/ecco_modules/cache")),
                 storage_options_yaml_file: str = str(here("s3_secret.yaml")),
                 disable_caching: bool = False,
                 image_train_transform: Callable[[
                     Image.Image], T_co] = transform_binary_PIL_image_to_bool_tensor,
                 image_test_transform: Callable[[
                     Image.Image], T_co] = transform_binary_PIL_image_to_bool_tensor,
                 text_train_transform: Callable[[
                     str], T2_co] = identity_transformation,
                 text_test_transform: Callable[[
                     str], T2_co] = identity_transformation,
                 **kwargs):
        self.storage_options = get_s3fs_credentials(storage_options_yaml_file)
        super().__init__(
            train_sqlite_url=train_sqlite_url,
            val_sqlite_url=val_sqlite_url,
            test_sqlite_url=test_sqlite_url,
            storage_options=self.storage_options,
            table_name="ecco_pages",
            index_column="entry_number",
            columns_to_return="img_offset, img_compressed_size + 128, txt_offset, txt_compressed_size + 128, id",
            id_column="id",
            cache_dir=cache_dir,
            **kwargs)
        self.disable_caching = disable_caching
        self.image_zip_url = image_zip_url
        self.image_zip_size = get_async_filesystem(
            image_zip_url, self.storage_options).size(image_zip_url)
        self.image_afetcher = PLocalAFetcher(
            image_zip_url, self.image_zip_size, storage_options=self.storage_options, cache_dir=cache_dir if not disable_caching else None)
        self.text_zip_url = text_zip_url
        self.text_zip_size = get_async_filesystem(
            text_zip_url, self.storage_options).size(text_zip_url)
        self.text_afetcher = PLocalAFetcher(
            text_zip_url, self.text_zip_size, storage_options=self.storage_options, cache_dir=cache_dir if not disable_caching else None)
        self.train_transform = functools.partial(_ecco_images_and_page_texts_transform, image_afetcher=self.image_afetcher,
                                                 text_afetcher=self.text_afetcher, image_transform=image_train_transform, text_transform=text_train_transform)  # type: ignore
        self.test_transform = functools.partial(_ecco_images_and_page_texts_transform, image_afetcher=self.image_afetcher,
                                                text_afetcher=self.text_afetcher, image_transform=image_test_transform, text_transform=text_test_transform)

    def prepare_data(self):
        # Ensure sqlite databases are downloaded
        super().prepare_data()
        # Start a background process to read the ZIP file into our mmap cache if zip_url is in S3
        if not self.disable_caching:
            prefetch_if_remote(self.image_zip_url, self.image_zip_size,
                               self.cache_dir, self.storage_options)
            prefetch_if_remote(self.text_zip_url, self.text_zip_size,
                               self.cache_dir, self.storage_options)


class ECCOImagesAndTextTokensDataModule(ECCOImagesAndPageTextsDataModule[torch.Tensor, str]):
    def __init__(self, model: str = "TurkuNLP/eccobert-base-cased-v1", **kwargs):
        self.model = model
        super().__init__(collate_fn=None, **kwargs)

    def prepare_data(self):
        super().prepare_data()
        # AutoTokenizer.from_pretrained(self.model) # Rust tokenizer complains about parallelism if we do this

    def setup(self, stage: str | None):
        if self.collate_fn is None:
            self.collate_fn = functools.partial(_batch_tokenise_after_collate, tokenizer=AutoTokenizer.from_pretrained(
                self.model), text_index=2, collate_fn=_remove_nones_from_batch)
        super().setup(stage)


class ECCOImagesAndTextEmbeddingsDataModule(ECCOImagesAndPageTextsDataModule[torch.Tensor, torch.Tensor]):
    def __init__(self, model: str = "TurkuNLP/eccobert-base-cased-v1", **kwargs):
        self.model = model
        super().__init__(collate_fn=None, **kwargs)

    def prepare_data(self):
        super().prepare_data()
        AutoTokenizer.from_pretrained(self.model)
        AutoModel.from_pretrained(self.model)

    def setup(self, stage: str | None):
        if self.collate_fn is None:
            self.collate_fn = functools.partial(_batch_embed_after_collate, tokenizer=AutoTokenizer.from_pretrained(
                self.model), model=AutoModel.from_pretrained(self.model), text_index=2, collate_fn=_remove_nones_from_batch)
        super().setup(stage)
